# URL:     http://www.fiber-space.de
# Author:  Kay Schluehr <easyextend@fiber-space.de>
# Date:    10 May 2006
import EasyExtend
from   EasyExtend.parser.DFAParser import DFAParser
from   EasyExtend.eetokenizer import StdTokenizer, LineList
import EasyExtend.csttools as csttools
import EasyExtend.util.path as path
import EasyExtend.fibers.zero.fiber as zero_fiber
import eecommon
import marshal
import imp
import struct
import os
from   EasyExtend.parser.PyParser import PyParser
import cst2source
import cst
import sys

PyCF_DONT_IMPLY_DEDENT = 0x200   # ??? some undocumented compiler flag with unkown functionality but effect.

class EEGrammar(eecommon.EEShow):
    def __init__(self, fiber):
        self.options    = fiber.options
        self.fiber      = fiber
        self.grammarObj = self.createGrammarObject()

    def createGrammarObject(self):
        '''
        Function used to create ext-language specific grammarObjects that contain
        dfa parser tables. These grammar objects will be created using lifted nodes of the
        the used fibers.

        @note: The grammarObj will be replaced by a newer version if the Grammar file
            is more recent than the parsetable.py module.
        '''
        parsetable   = self.fiber.parsetable
        f_parsetable = path.path(parsetable.__file__)
        f_parsetable = path.path(f_parsetable.replace("pyc","py"))

        # is parsetable.py file module more recent than the Grammar file or small than return the
        # grammarObj immediately ...
        f_parsetable_stat = os.stat(f_parsetable)
        gram_ext = self._grammar_ext()
        if gram_ext:
            if f_parsetable_stat[-2] >= os.stat(gram_ext)[-2]:
                return parsetable.grammarObj
            else:
                gram_lines = self._merge_grammar_ext(gram_ext)
        elif f_parsetable_stat[-2] >= os.stat(f_parsetable.dirname()+os.sep+"Grammar")[-2]:
                return parsetable.grammarObj

        # create a new grammarObj
        import EasyExtend.parser.PgenParser as PgenParser
        import EasyExtend.parser.PyPgen as PyPgen
        import EasyExtend.parser.PyToken as PyToken
        import EasyExtend.eetokenizer as tokenizer
        PyToken.FIBER_OFFSET = self.fiber.FIBER_OFFSET
        PyToken.FiberToken().gen_token(PyToken)
        if gram_ext:
            grammarST  = PgenParser.Parser(self.fiber).parseLines(gram_lines,
                                                                  tokenizer.StdTokenizer(PyToken, typ="grammar", use_comment=False))
        else:
            grammarST  = PgenParser.Parser(self.fiber).parseFile(f_parsetable.dirname()+os.sep+"Grammar",
                                                                 tokenizer.StdTokenizer(PyToken, typ="grammar", use_comment=False))
        pgenObj    = PyPgen.PyPgen(self.fiber)
        grammarObj = pgenObj.createGrammar(grammarST)
        fFSGrammar = open(f_parsetable,"w")
        print "*** Modify parsetable.py file ***"
        print >> fFSGrammar, "# %s" % ("_" * 70)
        print >> fFSGrammar, "# This was automatically generated by PyPgen."
        print >> fFSGrammar, "# Hack at your own risk."
        print >> fFSGrammar
        import pprint
        accelGrammarObj = DFAParser(self.fiber, **self.options).addAccelerators(grammarObj)
        print >> fFSGrammar, "grammarObj =",
        pprint.pprint(accelGrammarObj,fFSGrammar)
        fFSGrammar.close()
        reload(parsetable)
        reload(PyToken)
        fFSSymbol = open(os.sep.join([f_parsetable.dirname(),"fibermod","symbol.py"]),"w")
        self.mapSymbols(fFSSymbol, grammarObj)
        fFSSymbol.close()
        reload(self.fiber.symbol)
        return parsetable.grammarObj


    def _grammar_ext(self):
        FiberPath  = path.path(self.fiber.__file__).dirname()
        GrammarExt = FiberPath.joinpath("Grammar.ext")
        if GrammarExt.isfile():
            return GrammarExt

    def _merge_grammar_ext(self, GrammarExt):
        EEPath = path.path(EasyExtend.__file__).dirname()

        GrammarLines   = open(EEPath.joinpath("Grammar")).readlines()
        GrammarExtText = open(GrammarExt).read().strip()
        if GrammarExtText:
            GrammarExtLines = GrammarExtText.splitlines()
        else:
            return LineList(GrammarLines)

        def create_rules(lines):
            k = 1
            rule = ""
            rules = {}
            for line in lines:
                line = line.strip()
                if not line:
                    continue
                hidx = line.find("#")
                if hidx>=0:
                    line = line[:hidx]
                    if not line:
                        continue
                colidx = line.find(":")
                if colidx>=0:
                    rule_name = line[:colidx]
                    if not rule_name.split("_")[0].isalnum():
                        idx, rule_text = rules[rule]
                        rules[rule] = (idx, rule_text + " "+ line.strip())
                    else:
                        rule = rule_name
                        rules[rule] = (k, line)
                        k+=1
                else:
                    (idx, l) = rules[rule_name]
                    rules[rule_name] = (idx, l+line)
            return rules

        GrammarRules    = create_rules(GrammarLines)
        n = len(GrammarRules)+1
        GrammarExtRules = create_rules(GrammarExtLines)
        for r, v in GrammarExtRules.items():
            rule = GrammarRules.get(r)
            if rule:
                GrammarRules[r] = (rule[0], v[1])
            else:
                GrammarRules[r] = (n+v[0], v[1])

        grammar = [line for (k,line) in sorted(GrammarRules.values())]
        G = open("Grammar","w")
        for l in grammar:
            print >> G, l
        return LineList(grammar)


    def mapSymbols(self, fFSSymbol, grammarObj):
        '''
        A new language specific symbol.py file will be created.
        '''
        print >> fFSSymbol, "#  This file is automatically generated; please don't muck it up!"
        print >> fFSSymbol
        print >> fFSSymbol, "#--begin constants--"
        print >> fFSSymbol
        for tok in grammarObj[0]:
            print >> fFSSymbol, "%s = %s"%(tok[1],tok[0])
            #print "%s = %s"%(tok[1],tok[0])
        print >> fFSSymbol
        print >> fFSSymbol, "#--end constants--"
        print >> fFSSymbol
        print >> fFSSymbol, "sym_name = {}"
        print >> fFSSymbol, "for _name, _value in globals().items():"
        print >> fFSSymbol, "    if type(_value) is type(0):"
        print >> fFSSymbol, "        sym_name[_value] = _name"


class EECompiler(eecommon.EEShow):
    def __init__(self, fiber = zero_fiber):
        self.options    = fiber.options
        self.fiber      = fiber
        self.eegrammar  = EEGrammar(fiber)
        self.grammarObj = self.eegrammar.createGrammarObject()

    def eetransform(self, cst, **kwd):
        transformer  = self.fiber.FiberTransformer(self.fiber, **kwd)
        try:
            transformer.run(cst)
        finally:
            transformer.terminate()

    def eetokenize(self, source):
        '''
        @param source: text to be tokenized.
        @return: token stream.
        '''
        tokenizer   = StdTokenizer(self.fiber.token)
        tokenizer.tokenize_input(cst2source.SourceCode(source).readline)
        return tokenizer.tokenized()

    def eetokenize_string(self, source):
        tokenizer   = StdTokenizer(self.fiber.token)
        tokenizer.tokenize_string(source)
        return tokenizer.tokenized()


    def eeparse_file(self, filename):
        '''
        @param filename: file to be parsed.
        @param grammarObj: language specific grammarObj read from parsetable.py
        '''
        tokenizer = StdTokenizer(self.fiber.token)
        tokenizer.tokenize_file(filename)
        self.maybe_show_token("", filename = filename)
        parseTree = DFAParser(self.fiber, **self.options).parsetok(tokenizer, self.grammarObj, self.fiber.symbol.file_input)
        return parseTree

    def eeparse_source(self, source, start_symbol):
        tokenizer   = StdTokenizer(self.fiber.token)
        tokenizer.tokenize_input(cst2source.SourceCode(source).readline)
        cst = DFAParser(self.fiber, **self.options).parsetok(tokenizer, self.grammarObj, start_symbol)
        return cst

    def eeparse_expr(self, source):
        return self.eeparse_source(source, self.fiber.symbol.eval_input)

    def eeparse_suite(self, source):
        return self.eeparse_source(source, self.fiber.symbol.file_input)

    def eecompile_cst(self, node, **kwd):
        if node[0] != self.fiber.symbol.file_input:
            node = cst.file_input(node)
        try:
            return PyParser.tuple2ast(node).compile()
        except PyParser.ParserError:
            csttools.projection(node)
            return self.try_compile(node, **kwd)

    def eecompile_suite(self, source, **kwd):
        '''
        @param source: source in string form
        @param grammarObj: language specific grammarObj read from parsetable.py
        @param fiber: module object containing language definitions
        '''
        cst = self.eeparse_suite(source)
        self.eetransform(cst, **kwd)
        csttools.projection(cst)
        return self.eecompile_cst(cst)

    def eecompile_expr(self, source, **kwd):
        '''
        @param source: source in string form
        @param grammarObj: language specific grammarObj read from parsetable.py
        @param fiber: module object containing language definitions
        '''
        cst = self.eeparse_expr(source)
        self.eetransform(cst, **kwd)
        csttools.projection(cst)
        return self.try_compile(cst, **kwd)


    def try_compile(self, cst, **kwd):
        try:
            ast =  PyParser.tuple2ast(cst)
            return ast.compile()
        except PyParser.ParserError:
            #TBD: self.check_cst(cst, sys.exc_info()) <>
            src = self.fiber.unparse(cst)
            return compile(src,kwd.get("filename","<single>"),"exec", PyCF_DONT_IMPLY_DEDENT)

    def eecompile_file(self, filename, **kwd):
        '''
        Compile file according to a specific extension language into a Python module.

        @param filename: path to the destination file that will be compiled
        '''
        parseTree = self.eeparse_file(filename)
        mod = EEModule(parseTree, filename, self.fiber)
        kwd["options"]  = self.options
        kwd["filename"] = filename
        try:
            mod.compile(**kwd)
        except SyntaxError:
            raise
        else:
            try:
                ext = self.fiber.ext
            except AttributeError:
                ext = "pyc"
            idx = filename.rfind(".")+1
            f = open(filename[:idx] + ext, "wb")
            mod.dump(f)
            f.close()
            return f


class EEModule(eecommon.EEShow):
    '''
    Class used to represent a module object of the fiber.
    '''
    def __init__(self, tree, filename, fiber):
        '''
        @param tree: parse tree of the extension language
        @param filename: filename of the compiled module
        @param fiber: Python module object containing relevant ext-language definitions.
        '''
        self.tree = tree
        self.filename = filename
        self.code = None
        self.fiber = fiber

    mode = "exec"

    def preprocess(self, **kwd):
        self.options = kwd["options"]
        ori_token = []
        transformer = self.fiber.FiberTransformer(self.fiber, **kwd)
        self.maybe_show_cst_before(self.tree)
        transformer.run(self.tree)
        transformer.general_transform(self.tree)
        self.maybe_show_cst_after(self.tree)
        transformer.terminate()
        csttools.projection(self.tree)
        self.maybe_show_python(self.tree)
        if self.options["parse_only"]:
            print "File parsed -> exit"
            sys.exit(0)

    def compile(self, **kwd):
        '''
        Create Python code object from transformed extension language parse tree.
        '''
        self.preprocess(**kwd)
        try:
            ast = PyParser.tuple2ast(self.tree)      # usual compilation process
            self.code = ast.compile(self.filename)
        except PyParser.ParserError, e:
            source = self.fiber.unparse(self.tree)          #  parser module is buggy...
            self.code = compile(source, self.filename, "exec")

    def dump(self, f):
        '''
        Dump Python code-object into a binary file
        '''
        f.write(self.getPycHeader())
        marshal.dump(self.code, f)

    MAGIC = imp.get_magic()

    def getPycHeader(self):
        # compile.c uses marshal to write a long directly, with
        # calling the interface that would also generate a 1-byte code
        # to indicate the type of the value.  simplest way to get the
        # same effect is to call marshal and then skip the code.
        mtime = os.path.getmtime(self.filename)
        mtime = struct.pack('<i', mtime)
        return self.MAGIC + mtime


