from conf import*       # fiber specific modules ( symbol, token ) and objects ( prompt, pprint, unparse ...)
import EasyExtend       # EasyExtend package
import parsetable       # contains fiber specific DFA tables used by the parser
from EasyExtend.csttools import*   # tools used to manipulate CSTs
from EasyExtend.eetransformer import Transformer, transform

def transpath_expr(idx):
    return Transpath.paths[idx]

class Transpath(object):
    paths = {}
    def __init__(self, *args):
        self.args = args

class Transpath_Variable(Transpath):pass


class FiberTransformer(Transformer):
    '''
    Defines fiber specific transformations
    '''
    @transform
    def file_input(self, node):
        transpath = self.path_expr(find_node(node, symbol.path_expr))
        _id = id(transpath)
        Transpath.paths[_id] = transpath
        _stmt = any_stmt(any_test(CST_CallFunc("transpath_expr",[Number(_id)])))
        return file_input(*[_stmt])

    def path_expr(self, node):
        return Transpath(node)

    def file_pattern(self, node):
        tok = find_node(node, token.FAT_TOKEN, level=1)
        if tok:
            return any_expr(self.FAT_TOKEN(tok))
        else:
            return self.path_expr(find_node(node, symbol.path_expr))

    @transform
    def path_variable(self, node):
        return CST_CallFunc("TransPath_Variable",[self.FAT_TOKEN(node[1][1])])

    @transform
    def FAT_TOKEN(self, node):
        return String(node[1])

    @transform
    def separator(self, node):
        # verify
        if not find_node(node, token.Separator):
            raise SyntaxError("Path shall use path separator '%s'",token.TOKEN_MAP[token.Separator])

__publish__ = ["Transpath","transpath_expr"]

if __name__ == '__main__':
    import fiber
    (options,args) = opt.parse_args()
    if options.path_mode:
        if options.path_mode == 'native':
            fiber.token.change_path_mode("NATIVE")
        elif options.path_mode == 'common':
            fiber.token.change_path_mode("COMMON")
        else:
            raise ValueError("'%s' cannot be value of option --path_mode. Available options: native | common"%options.path_mode)
    if args:
        py_module = args[-1]
        EasyExtend.run_module(py_module, fiber, **options.__dict__)
    else:
        console = EasyExtend.create_console("sub", fiber, **options.__dict__)
        console.interact()
